# Q learning for the OpenAI Gym Taxi enviromnent
import gym
import numpy as np
import random

random.seed(42)
np.random.seed(42)

import warnings; warnings.filterwarnings('ignore', category=DeprecationWarning)

if __name__ == '__main__':
    """
    conda activate /home/wax/anaconda3/envs/reinforce_example
    """
    # Environment Setup:
    env = gym.make('Taxi-v3')
    env.reset()
    #env.render()

    # Q[state, action] table implementation for SARSA:
    #
    Q = np.zeros([env.observation_space.n, env.action_space.n])
    gamma = 0.7    # discount factor
    alpha = 0.2    # learning rate
    epsilon = 0.1  # epsilon greedy
    for episode in range(10000):
        done = False
        total_reward = 0
        current_state, _ = env.reset()
        if random.uniform(0, 1) < epsilon:
            current_action = env.action_space.sample() # explore state space
        else:
            current_action = np.argmax(Q[current_state]) # exploit learned values
        while not done:
            next_state, reward, terminated, truncated, info = env.step(current_action); done = terminated or truncated
            if random.uniform(0, 1) < epsilon:
                next_action = env.action_space.sample() # explore state space
            else:
                next_action = np.argmax(Q[next_state]) # exploit learned values

            sarsa_value = Q[next_state, next_action]
            old_value = Q[current_state, current_action]
            new_value = old_value + alpha * ( reward + gamma * sarsa_value - old_value )

            Q[current_state, current_action] = new_value
            total_reward += reward

            current_state = next_state
            current_action = next_action
        if episode % 1000 == 0:
            print(f'Episode {episode:5d} Total Reward: {total_reward:5d}')

    # Evaluate the optimal SARSA Taxi result:
    #
    total_timesteps, total_penalties, total_rewards = 0, 0, 0
    ep = 100
    for _ in range(ep):
        state, _ = env.reset()
        timesteps, rewards, penalties = 0, 0, 0
        done = False
        while not done:
            action = np.argmax(Q[state])
            state, reward, terminated, truncated, info = env.step(action); done = terminated or truncated
            timesteps += 1
            rewards += reward
            if reward == -10:
                penalties += 1
        total_timesteps += timesteps
        total_rewards += rewards
        total_penalties += penalties
    print(f'SARSA results after {ep} episodes')
    print(f'Average timesteps per episodes: {total_timesteps/ep:6.2f}')
    print(f'Average reward per episodes: {total_rewards/ep:6.2f}')
    #print(f'Average penalties per episodes: {total_penalties/ep:6.2f}')
